from __future__ import annotations

from typing import Literal

import pytest

from PIL import ContainerIO, Image

from .helper import hopper

TEST_FILE = "Tests/images/dummy.container"


def test_sanity() -> None:
    dir(Image)
    dir(ContainerIO)


def test_isatty() -> None:
    with hopper() as im:
        container = ContainerIO.ContainerIO(im, 0, 0)

    assert container.isatty() is False


@pytest.mark.parametrize(
    "mode, expected_position",
    (
        (0, 33),
        (1, 66),
        (2, 100),
    ),
)
def test_seek_mode(mode: Literal[0, 1, 2], expected_position: int) -> None:
    # Arrange
    with open(TEST_FILE, "rb") as fh:
        container = ContainerIO.ContainerIO(fh, 22, 100)

        # Act
        container.seek(33, mode)
        container.seek(33, mode)

        # Assert
        assert container.tell() == expected_position


@pytest.mark.parametrize("bytesmode", (True, False))
def test_read_n0(bytesmode: bool) -> None:
    # Arrange
    with open(TEST_FILE, "rb" if bytesmode else "r") as fh:
        container = ContainerIO.ContainerIO(fh, 22, 100)

        # Act
        container.seek(81)
        data = container.read()

        # Assert
        if bytesmode:
            data = data.decode()
        assert data == "7\nThis is line 8\n"


@pytest.mark.parametrize("bytesmode", (True, False))
def test_read_n(bytesmode: bool) -> None:
    # Arrange
    with open(TEST_FILE, "rb" if bytesmode else "r") as fh:
        container = ContainerIO.ContainerIO(fh, 22, 100)

        # Act
        container.seek(81)
        data = container.read(3)

        # Assert
        if bytesmode:
            data = data.decode()
        assert data == "7\nT"


@pytest.mark.parametrize("bytesmode", (True, False))
def test_read_eof(bytesmode: bool) -> None:
    # Arrange
    with open(TEST_FILE, "rb" if bytesmode else "r") as fh:
        container = ContainerIO.ContainerIO(fh, 22, 100)

        # Act
        container.seek(100)
        data = container.read()

        # Assert
        if bytesmode:
            data = data.decode()
        assert data == ""


@pytest.mark.parametrize("bytesmode", (True, False))
def test_readline(bytesmode: bool) -> None:
    # Arrange
    with open(TEST_FILE, "rb" if bytesmode else "r") as fh:
        container = ContainerIO.ContainerIO(fh, 0, 120)

        # Act
        data = container.readline()

        # Assert
        if bytesmode:
            data = data.decode()
        assert data == "This is line 1\n"


@pytest.mark.parametrize("bytesmode", (True, False))
def test_readlines(bytesmode: bool) -> None:
    # Arrange
    expected = [
        "This is line 1\n",
        "This is line 2\n",
        "This is line 3\n",
        "This is line 4\n",
        "This is line 5\n",
        "This is line 6\n",
        "This is line 7\n",
        "This is line 8\n",
    ]
    with open(TEST_FILE, "rb" if bytesmode else "r") as fh:
        container = ContainerIO.ContainerIO(fh, 0, 120)

        # Act
        data = container.readlines()

        # Assert
        if bytesmode:
            data = [line.decode() for line in data]
        assert data == expected



from __future__ import annotations

from .helper import hopper


def test_sanity() -> None:
    im = hopper()
    type_repr = repr(type(im.getim()))

    assert "PyCapsule" in type_repr
    assert isinstance(im.im.id, int)



from __future__ import annotations

import pytest

from PIL import Image, ImageMath


def pixel(im: Image.Image | int) -> str | int:
    if isinstance(im, int):
        return int(im)  # hack to deal with booleans

    return f"{im.mode} {repr(im.getpixel((0, 0)))}"


A = Image.new("L", (1, 1), 1)
B = Image.new("L", (1, 1), 2)
Z = Image.new("L", (1, 1), 0)  # Z for zero
F = Image.new("F", (1, 1), 3)
I = Image.new("I", (1, 1), 4)  # noqa: E741

A2 = A.resize((2, 2))
B2 = B.resize((2, 2))

images = {"A": A, "B": B, "F": F, "I": I}


def test_sanity() -> None:
    assert ImageMath.unsafe_eval("1") == 1
    assert ImageMath.unsafe_eval("1+A", A=2) == 3
    assert pixel(ImageMath.unsafe_eval("A+B", A=A, B=B)) == "I 3"
    assert pixel(ImageMath.unsafe_eval("A+B", images)) == "I 3"
    assert pixel(ImageMath.unsafe_eval("float(A)+B", images)) == "F 3.0"
    assert pixel(ImageMath.unsafe_eval("int(float(A)+B)", images)) == "I 3"


def test_eval_deprecated() -> None:
    with pytest.warns(DeprecationWarning):
        assert ImageMath.eval("1") == 1


def test_ops() -> None:
    assert pixel(ImageMath.unsafe_eval("-A", images)) == "I -1"
    assert pixel(ImageMath.unsafe_eval("+B", images)) == "L 2"

    assert pixel(ImageMath.unsafe_eval("A+B", images)) == "I 3"
    assert pixel(ImageMath.unsafe_eval("A-B", images)) == "I -1"
    assert pixel(ImageMath.unsafe_eval("A*B", images)) == "I 2"
    assert pixel(ImageMath.unsafe_eval("A/B", images)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("B**2", images)) == "I 4"
    assert pixel(ImageMath.unsafe_eval("B**33", images)) == "I 2147483647"

    assert pixel(ImageMath.unsafe_eval("float(A)+B", images)) == "F 3.0"
    assert pixel(ImageMath.unsafe_eval("float(A)-B", images)) == "F -1.0"
    assert pixel(ImageMath.unsafe_eval("float(A)*B", images)) == "F 2.0"
    assert pixel(ImageMath.unsafe_eval("float(A)/B", images)) == "F 0.5"
    assert pixel(ImageMath.unsafe_eval("float(B)**2", images)) == "F 4.0"
    assert pixel(ImageMath.unsafe_eval("float(B)**33", images)) == "F 8589934592.0"


@pytest.mark.parametrize(
    "expression",
    (
        "exec('pass')",
        "(lambda: exec('pass'))()",
        "(lambda: (lambda: exec('pass'))())()",
    ),
)
def test_prevent_exec(expression: str) -> None:
    with pytest.raises(ValueError):
        ImageMath.unsafe_eval(expression)


def test_prevent_double_underscores() -> None:
    with pytest.raises(ValueError):
        ImageMath.unsafe_eval("1", {"__": None})


def test_prevent_builtins() -> None:
    with pytest.raises(ValueError):
        ImageMath.unsafe_eval("(lambda: exec('exit()'))()", {"exec": None})


def test_logical() -> None:
    assert pixel(ImageMath.unsafe_eval("not A", images)) == 0
    assert pixel(ImageMath.unsafe_eval("A and B", images)) == "L 2"
    assert pixel(ImageMath.unsafe_eval("A or B", images)) == "L 1"


def test_convert() -> None:
    assert pixel(ImageMath.unsafe_eval("convert(A+B, 'L')", images)) == "L 3"
    assert pixel(ImageMath.unsafe_eval("convert(A+B, '1')", images)) == "1 0"
    assert (
        pixel(ImageMath.unsafe_eval("convert(A+B, 'RGB')", images)) == "RGB (3, 3, 3)"
    )


def test_compare() -> None:
    assert pixel(ImageMath.unsafe_eval("min(A, B)", images)) == "I 1"
    assert pixel(ImageMath.unsafe_eval("max(A, B)", images)) == "I 2"
    assert pixel(ImageMath.unsafe_eval("A == 1", images)) == "I 1"
    assert pixel(ImageMath.unsafe_eval("A == 2", images)) == "I 0"


def test_one_image_larger() -> None:
    assert pixel(ImageMath.unsafe_eval("A+B", A=A2, B=B)) == "I 3"
    assert pixel(ImageMath.unsafe_eval("A+B", A=A, B=B2)) == "I 3"


def test_abs() -> None:
    assert pixel(ImageMath.unsafe_eval("abs(A)", A=A)) == "I 1"
    assert pixel(ImageMath.unsafe_eval("abs(B)", B=B)) == "I 2"


def test_binary_mod() -> None:
    assert pixel(ImageMath.unsafe_eval("A%A", A=A)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("B%B", B=B)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("A%B", A=A, B=B)) == "I 1"
    assert pixel(ImageMath.unsafe_eval("B%A", A=A, B=B)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("Z%A", A=A, Z=Z)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("Z%B", B=B, Z=Z)) == "I 0"


def test_bitwise_invert() -> None:
    assert pixel(ImageMath.unsafe_eval("~Z", Z=Z)) == "I -1"
    assert pixel(ImageMath.unsafe_eval("~A", A=A)) == "I -2"
    assert pixel(ImageMath.unsafe_eval("~B", B=B)) == "I -3"


def test_bitwise_and() -> None:
    assert pixel(ImageMath.unsafe_eval("Z&Z", A=A, Z=Z)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("Z&A", A=A, Z=Z)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("A&Z", A=A, Z=Z)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("A&A", A=A, Z=Z)) == "I 1"


def test_bitwise_or() -> None:
    assert pixel(ImageMath.unsafe_eval("Z|Z", A=A, Z=Z)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("Z|A", A=A, Z=Z)) == "I 1"
    assert pixel(ImageMath.unsafe_eval("A|Z", A=A, Z=Z)) == "I 1"
    assert pixel(ImageMath.unsafe_eval("A|A", A=A, Z=Z)) == "I 1"


def test_bitwise_xor() -> None:
    assert pixel(ImageMath.unsafe_eval("Z^Z", A=A, Z=Z)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("Z^A", A=A, Z=Z)) == "I 1"
    assert pixel(ImageMath.unsafe_eval("A^Z", A=A, Z=Z)) == "I 1"
    assert pixel(ImageMath.unsafe_eval("A^A", A=A, Z=Z)) == "I 0"


def test_bitwise_leftshift() -> None:
    assert pixel(ImageMath.unsafe_eval("Z<<0", Z=Z)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("Z<<1", Z=Z)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("A<<0", A=A)) == "I 1"
    assert pixel(ImageMath.unsafe_eval("A<<1", A=A)) == "I 2"


def test_bitwise_rightshift() -> None:
    assert pixel(ImageMath.unsafe_eval("Z>>0", Z=Z)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("Z>>1", Z=Z)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("A>>0", A=A)) == "I 1"
    assert pixel(ImageMath.unsafe_eval("A>>1", A=A)) == "I 0"


def test_logical_eq() -> None:
    assert pixel(ImageMath.unsafe_eval("A==A", A=A)) == "I 1"
    assert pixel(ImageMath.unsafe_eval("B==B", B=B)) == "I 1"
    assert pixel(ImageMath.unsafe_eval("A==B", A=A, B=B)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("B==A", A=A, B=B)) == "I 0"


def test_logical_ne() -> None:
    assert pixel(ImageMath.unsafe_eval("A!=A", A=A)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("B!=B", B=B)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("A!=B", A=A, B=B)) == "I 1"
    assert pixel(ImageMath.unsafe_eval("B!=A", A=A, B=B)) == "I 1"


def test_logical_lt() -> None:
    assert pixel(ImageMath.unsafe_eval("A<A", A=A)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("B<B", B=B)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("A<B", A=A, B=B)) == "I 1"
    assert pixel(ImageMath.unsafe_eval("B<A", A=A, B=B)) == "I 0"


def test_logical_le() -> None:
    assert pixel(ImageMath.unsafe_eval("A<=A", A=A)) == "I 1"
    assert pixel(ImageMath.unsafe_eval("B<=B", B=B)) == "I 1"
    assert pixel(ImageMath.unsafe_eval("A<=B", A=A, B=B)) == "I 1"
    assert pixel(ImageMath.unsafe_eval("B<=A", A=A, B=B)) == "I 0"


def test_logical_gt() -> None:
    assert pixel(ImageMath.unsafe_eval("A>A", A=A)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("B>B", B=B)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("A>B", A=A, B=B)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("B>A", A=A, B=B)) == "I 1"


def test_logical_ge() -> None:
    assert pixel(ImageMath.unsafe_eval("A>=A", A=A)) == "I 1"
    assert pixel(ImageMath.unsafe_eval("B>=B", B=B)) == "I 1"
    assert pixel(ImageMath.unsafe_eval("A>=B", A=A, B=B)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("B>=A", A=A, B=B)) == "I 1"


def test_logical_equal() -> None:
    assert pixel(ImageMath.unsafe_eval("equal(A, A)", A=A)) == "I 1"
    assert pixel(ImageMath.unsafe_eval("equal(B, B)", B=B)) == "I 1"
    assert pixel(ImageMath.unsafe_eval("equal(Z, Z)", Z=Z)) == "I 1"
    assert pixel(ImageMath.unsafe_eval("equal(A, B)", A=A, B=B)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("equal(B, A)", A=A, B=B)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("equal(A, Z)", A=A, Z=Z)) == "I 0"


def test_logical_not_equal() -> None:
    assert pixel(ImageMath.unsafe_eval("notequal(A, A)", A=A)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("notequal(B, B)", B=B)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("notequal(Z, Z)", Z=Z)) == "I 0"
    assert pixel(ImageMath.unsafe_eval("notequal(A, B)", A=A, B=B)) == "I 1"
    assert pixel(ImageMath.unsafe_eval("notequal(B, A)", A=A, B=B)) == "I 1"
    assert pixel(ImageMath.unsafe_eval("notequal(A, Z)", A=A, Z=Z)) == "I 1"



from __future__ import annotations

import time

from PIL import PyAccess

from .helper import hopper

# Not running this test by default. No DOS against CI.


def iterate_get(size, access) -> None:
    (w, h) = size
    for x in range(w):
        for y in range(h):
            access[(x, y)]


def iterate_set(size, access) -> None:
    (w, h) = size
    for x in range(w):
        for y in range(h):
            access[(x, y)] = (x % 256, y % 256, 0)


def timer(func, label, *args) -> None:
    iterations = 5000
    starttime = time.time()
    for x in range(iterations):
        func(*args)
        if time.time() - starttime > 10:
            break
    endtime = time.time()
    print(
        "{}: completed {} iterations in {:.4f}s, {:.6f}s per iteration".format(
            label, x + 1, endtime - starttime, (endtime - starttime) / (x + 1.0)
        )
    )


def test_direct() -> None:
    im = hopper()
    im.load()
    # im = Image.new("RGB", (2000, 2000), (1, 3, 2))
    caccess = im.im.pixel_access(False)
    access = PyAccess.new(im, False)

    assert caccess[(0, 0)] == access[(0, 0)]

    print(f"Size: {im.width}x{im.height}")
    timer(iterate_get, "PyAccess - get", im.size, access)
    timer(iterate_set, "PyAccess - set", im.size, access)
    timer(iterate_get, "C-api - get", im.size, caccess)
    timer(iterate_set, "C-api - set", im.size, caccess)
